/*
 *  partbl_amiga.c -- Amiga RDSK partition table parser
 *
 *  Code extracted from
 *  linux/drivers/block/genhd.c
 *
 */

#include <linux/module.h>

#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/affs_hardblocks.h>
#include <asm/byteorder.h>

static __inline__ u32
checksum_block(u32 *m, int size)
{
	u32 sum = 0;

	while (size--)
		sum += be32_to_cpu(*m++);
	return sum;
}

static int
amiga_partition(struct gendisk *hd, kdev_t dev, unsigned long first_sector)
{
	struct buffer_head	*bh;
	struct RigidDiskBlock	*rdb;
	struct PartitionBlock	*pb;
	int			 start_sect;
	int			 nr_sects;
	int			 blk;
	int			 part, res;

	MOD_INC_USE_COUNT;
	set_blocksize(dev,512);
	res = 0;

	for (blk = 0; blk < RDB_ALLOCATION_LIMIT; blk++) {
		if(!(bh = bread(dev,blk,512))) {
			printk("Dev %s: unable to read RDB block %d\n",
			       kdevname(dev),blk);
			res = -1;
			goto rdb_done;
		}
		if (*(u32 *)bh->b_data == cpu_to_be32(IDNAME_RIGIDDISK)) {
			rdb = (struct RigidDiskBlock *)bh->b_data;
			if (checksum_block((u32 *)bh->b_data,be32_to_cpu(rdb->rdb_SummedLongs) & 0x7F)) {
				printk("Dev %s: RDB in block %d has bad checksum\n",
				       kdevname(dev),blk);
				brelse(bh);
				continue;
			}
			printk(" RDSK");
			blk = be32_to_cpu(rdb->rdb_PartitionList);
			brelse(bh);
			for (part = 1; blk > 0 && part <= 16; part++) {
				if (!(bh = bread(dev,blk,512))) {
					printk("Dev %s: unable to read partition block %d\n",
						       kdevname(dev),blk);
					res = -1;
					goto rdb_done;
				}
				pb  = (struct PartitionBlock *)bh->b_data;
				blk = be32_to_cpu(pb->pb_Next);
				if (pb->pb_ID == cpu_to_be32(IDNAME_PARTITION) && checksum_block(
				    (u32 *)pb,be32_to_cpu(pb->pb_SummedLongs) & 0x7F) == 0 ) {
					
					/* Tell Kernel about it */

					if (!(nr_sects = (be32_to_cpu(pb->pb_Environment[10]) + 1 -
							  be32_to_cpu(pb->pb_Environment[9])) *
							 be32_to_cpu(pb->pb_Environment[3]) *
							 be32_to_cpu(pb->pb_Environment[5]))) {
						continue;
					}
					start_sect = be32_to_cpu(pb->pb_Environment[9]) *
						     be32_to_cpu(pb->pb_Environment[3]) *
						     be32_to_cpu(pb->pb_Environment[5]);
					add_partition(hd,current_minor,start_sect,nr_sects);
					current_minor++;
					res = 1;
				}
				brelse(bh);
			}
			printk("\n");
			break;
		}
	}

rdb_done:
	set_blocksize(dev,BLOCK_SIZE);
	MOD_DEC_USE_COUNT;
	return res;
}

static struct partition_parser amiga_partition_parser =
    { NULL, "amiga", amiga_partition };

#ifndef MODULE
__initfunc(void amiga_partbl_init( void ))
{
	(void)register_partbl( &amiga_partition_parser );
}
#else
int init_module(void)
{
	return register_partbl( &amiga_partition_parser );
}

void cleanup_module(void)
{
	(void)unregister_partbl( &amiga_partition_parser );
}
#endif
